import logging
from typing import Dict, Any, Optional, List
from cota.knowledge.knowledge import Knowledge
from cota.constant import DEFAULT_DIALOGUE_MAX_TOKENS

logger = logging.getLogger(__name__)


class LLMKnowledge(Knowledge):
    """LLM-based knowledge retrieval using language models."""
    
    def __init__(self, config: Dict[str, Any]):
        super().__init__(config)
        # Support both single llm and llms list configuration
        self.llms = config.get('llms', [])
        self.prompt_template = config.get('prompt', self._default_prompt())
        self.max_tokens = config.get('max_tokens', DEFAULT_DIALOGUE_MAX_TOKENS)
        self.temperature = config.get('temperature', 0.7)
        
        if not self.llms:
            raise ValueError(f"LLM configuration is required for LLMKnowledge")
        
        # Build action to LLM mapping and find default LLM
        self.action_llm_map = {}
        self.default_llm = None
        
        for llm_config in self.llms:
            llm_name = llm_config.get('name')
            action = llm_config.get('action')
            
            if not llm_name:
                logger.warning(f"LLM configuration missing name: {llm_config}")
                continue
                
            if action:
                self.action_llm_map[action] = llm_name
            else:
                # LLM without specific action serves as default
                self.default_llm = llm_name
    
    def _default_prompt(self) -> str:
        """Default prompt template for knowledge retrieval."""
        return """根据以下查询问题，提供相关的知识信息：
                查询问题: {{query}}

                对话历史:
                {{history}}

                请提供准确、有用的知识信息来帮助回答这个问题。
            """
    
    def _select_llm(self, context: Dict[str, Any] = None) -> str:
        """Select appropriate LLM based on context."""
        if context:
            # Try to get current action from context
            current_action = context.get('current_action')
            if current_action and hasattr(current_action, 'name'):
                action_name = current_action.name
            elif current_action and hasattr(current_action, '__class__'):
                action_name = current_action.__class__.__name__
            else:
                action_name = None
                
            # Check if we have a specific LLM for this action
            if action_name and action_name in self.action_llm_map:
                return self.action_llm_map[action_name]
        
        # Use default LLM if available
        if self.default_llm:
            return self.default_llm
            
        # Fallback to first available LLM
        if self.llms:
            return self.llms[0].get('name')
            
        raise ValueError("No LLM available for knowledge retrieval")
    
    async def retrieve(self, query: str, context: Dict[str, Any] = None) -> str:
        """Use LLM to generate knowledge based on query and context."""
        try:
            # Get agent from context
            agent = context.get('agent') if context else None
            if not agent:
                logger.error("Agent not provided in context for LLMKnowledge")
                return ""
            
            # Build prompt with query and context
            prompt = self._build_prompt(query, context)
            
            # Select appropriate LLM
            selected_llm = self._select_llm(context)
            
            # Call LLM
            llm_instance = agent.llm_instance(selected_llm)
            result = await llm_instance.generate_chat(
                messages=[{"role": "user", "content": prompt}],
                max_tokens=self.max_tokens,
                temperature=self.temperature
            )
            
            return result["content"]
            
        except Exception as e:
            logger.error(f"LLM knowledge retrieval failed: {e}")
            return ""
    
    def _build_prompt(self, query: str, context: Dict[str, Any] = None) -> str:
        """Build prompt from template, query and context."""
        prompt = self.prompt_template.replace('{{query}}', query)
        
        if context:
            # Add conversation history if available
            dst = context.get('dst')
            if dst:
                try:
                    history = "\n".join([
                        f"{msg.get('role', 'unknown')}: {msg.get('content', '')}"
                        for msg in dst.extract_messages()[-5:]  # Last 5 messages
                    ])
                    prompt = prompt.replace('{{history}}', history)
                except Exception as e:
                    logger.warning(f"Failed to extract history: {e}")
                    prompt = prompt.replace('{{history}}', "")
            else:
                prompt = prompt.replace('{{history}}', "")
        
        return prompt



import logging
from typing import Dict, Text, Tuple, Any
from .base import Executor
from cota.utils.http import HttpClientManager

logger = logging.getLogger(__name__)

class HttpExecutor(Executor):
    """HTTP executor
    
    Handles execution logic based on HTTP requests.
    """
    
    async def execute(self, data: Dict[Text, Any]) -> Tuple[Text, Dict]:
        """Execute HTTP request
        
        Args:
            data: Request data
            
        Returns:
            Tuple[Response text, Metadata]
        """
        url = self.config.get("url", "")
        method = self.config.get("method", "get").upper()
        
        try:
            # Get HTTP client
            client = await HttpClientManager.instance().get_client(
                self.config.get("client_key", "default")
            )
            
            # Execute request
            if method == "GET":
                response = await client.get(url, params=data)
            else:
                response = await client.request(method, url, json_data=data)
            
            if not response.ok:
                raise Exception(f"Request failed with status {response.status}")
                
            return response.data.get("text", ""), response.data.get("metadata", {})
        except Exception as e:
            logger.error(f"Failed to execute HTTP request: {str(e)}")
            return "", {"error": str(e)} 


"""Custom HTTP LLM client."""

import aiohttp
import logging
from typing import List, Dict, Text, Optional, Any, Union

from .base import LLMClient

logger = logging.getLogger(__name__)


class CustomHttpClient(LLMClient):
    """Custom HTTP client that passes all config parameters to the HTTP endpoint.
    
    This client provides maximum flexibility by forwarding all configuration
    parameters to the HTTP endpoint, allowing the endpoint to decide how to
    handle RAG, authentication, and other custom features.
    """
    
    def __init__(self, api_key: Text, base_url: Text, model: Text, **kwargs):
        self.api_key = api_key
        self.base_url = base_url
        self.model = model
        # Store all additional config parameters for flexible passing to HTTP endpoint
        self.extra_config = kwargs
        self.session = aiohttp.ClientSession()

    async def generate_chat(
        self, 
        messages: List[Dict[Text, Text]], 
        max_tokens: int = 500,
        response_format: Dict[Text, Text] = {"type": "text"},
        tools: Optional[List[Dict[Text, Any]]] = None,
        tool_choice: Optional[str] = None
    ) -> Dict[Text, Any]:
        headers = {
            'Authorization': f'Bearer {self.api_key}',
            'Content-Type': 'application/json'
        }
        
        # Build request data with core parameters
        data = {
            "model": self.model,
            "messages": messages,
            "max_tokens": max_tokens,
            "response_format": response_format
        }
        
        # Add all extra config parameters (knowledge_id, rag_prompt, etc.)
        # Let the HTTP endpoint decide how to handle them
        data.update(self.extra_config)
        
        # If tools are provided, add them to request data
        if tools:
            data["tools"] = tools
            if tool_choice:
                data["tool_choice"] = tool_choice
        
        async with self.session.post(self.base_url, json=data, headers=headers) as response:
            response.raise_for_status()
            response_data = await response.json()
            
            # Ensure consistent return format - if response is a string, wrap it in a dict
            if isinstance(response_data, str):
                return {"content": response_data}
            elif isinstance(response_data, dict) and "content" not in response_data:
                # If it's a dict but doesn't have content field, assume the whole response is content
                return {"content": str(response_data)}
            else:
                # If it's already a dict with content field, return as is
                return response_data

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.session.close()
    def __del__(self):
        if hasattr(self, 'session') and self.session and not self.session.closed:
            # Schedule session cleanup if event loop is available
            try:
                import asyncio
                loop = asyncio.get_event_loop()
                if loop.is_running():
                    loop.create_task(self.session.close())
            except:
                pass




from typing import Dict, Text, Tuple, Any
from .base import BasePlugin

class Plugin(BasePlugin):
    """Example plugin
    
    Demonstrates how to implement a custom plugin. This plugin implements a simple text processing function.
    """
    
    def __init__(self, config: Dict[Text, Any]) -> None:
        super().__init__(config)
        self.prefix = config.get("prefix", "")
        self.suffix = config.get("suffix", "")
        
    async def execute(self, data: Dict[Text, Any]) -> Tuple[Text, Dict]:
        """Execute plugin logic
        
        This example plugin adds prefix and suffix to input text.
        
        Args:
            data: Data dictionary containing "text" field
            
        Returns:
            Tuple[Processed text, Metadata]
        """
        input_text = data.get("text", "")
        processed_text = f"{self.prefix}{input_text}{self.suffix}"
        
        metadata = {
            "original_length": len(input_text),
            "processed_length": len(processed_text),
            "prefix_used": self.prefix,
            "suffix_used": self.suffix
        }
        
        return processed_text, metadata 


import copy
import logging
from typing import Optional, List, Dict, Text, Any
from cota.channels.channel import Channel
from cota.actions.action import Action
from cota.actions.user_utter import UserUtter
from cota.actions.bot_utter import BotUtter
from cota.actions.form import Form
from cota.message.message import Message
from cota.store import Store
from cota.dst import DST

logger = logging.getLogger(__name__)

class Processor:
    def __init__(
            self,
            agent: 'Agent',
            store: Store
    ):
        self.agent = agent
        self.store = store
        self.dst = None  # todo

    def _get_dst(self):
        pass

    def _update_dst(self):
        pass

    async def handle_message(
            self,
            message: Message,
            channel: Optional[Channel] = None
    ):
        """handle message"""
        if self.agent.dialogue.get('use_proxy_user') == True:
            await self._handle_message_proxy(message, channel)
            return

        action = Action.build_from_name(name='UserUtter')
        action.run_from_dict({
            "result": [message.as_dict()],
            "sender": message.sender or 'user',
            "sender_id": message.sender_id or 'default_user',
            "receiver": message.receiver or 'bot',
            "receiver_id": message.receiver_id or 'default_bot'
        })

        if channel:
            await self.execute_channel_effects(action, message.session_id, channel)

        self.dst = await self.get_tracker(message.session_id)
        self.dst.update(action)

        if message.receiver != 'bot':
            await self.save_tracker(self.dst)
            return

        await self._handle_bot_actions(message.session_id, channel)
        await self.save_tracker(self.dst)

    # TODO: Check if this is reasonable
    async def handle_session(self, session_id:Text, channel: Optional[Channel] = None):
        self.dst = await self.get_tracker(session_id)
        await self._handle_bot_actions(session_id, channel)
        await self.save_tracker(self.dst)

    async def _handle_message_proxy(self, message: Message, channel: Optional[Channel] = None):
        self.dst = await self.get_tracker(message.session_id)
        # user = message.metadata.get('user') or self.agent.user

        max_proxy_step = self.agent.dialogue.get('max_proxy_step')                
        for i in range(max_proxy_step):
            action_config = self.agent.actions.get('UserUtter', {})
            # Use default values when UserUtter is not configured in agent.yml
            from cota.constant import DEFAULT_QUERY_DESCRIPTION, DEFAULT_QUERY_PROMPT
            action = Action.build_from_name(
                name='UserUtter',
                description=action_config.get("description", DEFAULT_QUERY_DESCRIPTION),
                prompt=action_config.get("prompt", DEFAULT_QUERY_PROMPT)
            )
            # await action.run(self.agent, self.dst, user=user)
            await action.run(self.agent, self.dst)
            self.dst.update(action)
            
            # Output user's message to channel first
            if channel:
                await self.execute_channel_effects(action, message.session_id, channel)
            
            # Check if user wants to stop the conversation based on state field in JSON response
            if action.result and action.result[0].get('state', 'continue') == 'stop':
                logger.debug(f"Conversation ending - Final DST state: \n {self.dst.current_state()}")
                return 

            await self._handle_bot_actions(message.session_id, channel)
        await self.save_tracker(self.dst)

    async def _handle_bot_actions(self, session_id: Text, channel: Optional[Channel] = None):
        while True:
            bot_actions = await self.agent.generate_actions(self.dst)
            for action_item in bot_actions:
                # All actions are single actions now (no tuple handling)
                await action_item.run(self.agent, self.dst)
                self.dst.update(action_item)
                logger.debug(f"After DST updated: \n {self.dst.current_state()}")
                if channel:
                    await self.execute_channel_effects(action_item, session_id, channel)
                if isinstance(action_item, BotUtter):
                    return
            if len(bot_actions) > 1 and isinstance(bot_actions[-1], Form) and isinstance(bot_actions[-2], BotUtter):
                break

    async def get_tracker(
            self, session_id: Text
    ) -> Optional[DST]:
        """Get tracker based on session_id"""
        actions_dict = await self.store.retrieve(session_id)
        if actions_dict is None:
            return DST(session_id=session_id, agent=self.agent)
        else:
            tracker = DST.from_dict(
                dst_dict={ "session_id": session_id, "actions": actions_dict}, 
                agent=self.agent
            )
            return tracker

    async def save_tracker(self, tracker: DST) -> None:
        """Save tracker to tracker store"""
        await self.store.save(tracker)

    async def execute_channel_effects(
            self,
            action: Action,
            session_id: Text,
            channel: Channel,
    ) -> None:
        if isinstance(action, UserUtter):
            for res in action.result:
                await channel.send_response(session_id, copy.deepcopy(res))
        elif isinstance(action, BotUtter):
            for res in action.result:
                await channel.send_response(session_id, copy.deepcopy(res))


import json
import re

def extract_json_from_string(input_string):
    """
    Extract JSON strings from input string.

    :param input_string: Input string
    :return: List of extracted JSON strings
    """
    json_regex = r'\{[^{}]*\}'
    matches = re.findall(json_regex, input_string)

    valid_jsons = []
    for match in matches:
        try:
            # todo: handle invalid json without ""
            match = match.replace("'", '"')
            json_object = json.loads(match)
            valid_jsons.append(json_object)
        except json.JSONDecodeError:
            continue
    return valid_jsons

def extract_array_from_string(input_string):
    """
    Extract JSON Array from input string.

    :param input_string: Input string
    :return: Extracted JSON Array
    """
    # Use regex to match possible JSON arrays
    start_index = input_string.find('[')
    end_index = input_string.rfind(']') + 1
    json_array_str = input_string[start_index:end_index]
    try:
        json_array = json.loads(json_array_str)
    except json.JSONDecodeError:
        return []
    return json_array

def extract_action_from_string(input_string, actions):
    selected_actions = []
    for name in actions:
        if name in input_string:
            selected_actions.append(name)
    return selected_actions

def parser_text_with_slots(text):
    # Define regex pattern
    pattern = r'\{\{\s*(.*?)\s*\|\s*(.*?)\s*\}\}'
    
    # Find all matching slot names and values
    matches = re.findall(pattern, text)
    
    # Create dictionary of slot names and values
    slots = {slot_name: slot_value for slot_value, slot_name in matches}
    
    # Define replacement function to replace matched parts with slot names
    def replace_with_slot_name(match):
        return match.group(1)
    
    # Replace all {{slot_name | slot_value}} with slot_name in text
    restored_text = re.sub(pattern, replace_with_slot_name, text)
    
    # Return dictionary and restored text
    return slots, restored_text


import logging
import asyncio
import json
from typing import Dict, Text, Tuple, Any, Optional, List
from .base import Executor
from cota.llm import LLM

logger = logging.getLogger(__name__)

try:
    from mcp import ClientSession, StdioServerParameters
    from mcp.client.stdio import stdio_client
    MCP_AVAILABLE = True
except ImportError:
    MCP_AVAILABLE = False
    logger.warning(
        "MCP SDK not available. Please install it with: pip install mcp"
    )


class MCPClientWrapper:
    """MCP客户端包装类，管理MCP服务器连接和工具调用"""
    
    def __init__(self, server_params: StdioServerParameters):
        """
        初始化MCP客户端包装器
        
        Args:
            server_params: MCP服务器参数
        """
        self.server_params = server_params
        self._stdio_context = None
        self._read = None
        self._write = None
        self._session: Optional[ClientSession] = None
        self._initialized = False
        
    async def ensure_connected(self) -> ClientSession:
        """确保MCP连接已建立并初始化"""
        if not self._initialized or self._session is None:
            try:
                # 创建stdio客户端上下文管理器
                self._stdio_context = stdio_client(self.server_params)
                self._read, self._write = await self._stdio_context.__aenter__()
                
                # 创建会话
                self._session = ClientSession(self._read, self._write)
                
                # 初始化会话
                await self._session.initialize()
                self._initialized = True
                logger.info("MCP会话初始化成功")
                
            except Exception as e:
                logger.error(f"MCP连接失败: {e}")
                self._initialized = False
                raise
        
        return self._session
    
    async def list_tools(self) -> List[Dict]:
        """列出所有可用的工具"""
        session = await self.ensure_connected()
        try:
            result = await session.list_tools()
            tools = []
            for tool in result.tools:
                tools.append({
                    "name": tool.name,
                    "description": tool.description or "",
                    "inputSchema": tool.inputSchema if hasattr(tool, 'inputSchema') else {}
                })
            return tools
        except Exception as e:
            logger.error(f"获取工具列表失败: {e}")
            raise
    
    async def list_prompts(self) -> List[Dict]:
        """列出所有可用的prompts"""
        session = await self.ensure_connected()
        try:
            result = await session.list_prompts()
            prompts = []
            for prompt in result.prompts:
                prompts.append({
                    "name": prompt.name,
                    "description": prompt.description or "",
                    "arguments": prompt.arguments if hasattr(prompt, 'arguments') else []
                })
            return prompts
        except Exception as e:
            logger.error(f"获取prompts列表失败: {e}")
            raise
    
    async def call_tool(self, tool_name: str, arguments: Dict) -> Dict:
        """调用MCP工具"""
        session = await self.ensure_connected()
        try:
            result = await session.call_tool(tool_name, arguments)
            return {
                "content": result.content if hasattr(result, 'content') else [],
                "isError": result.isError if hasattr(result, 'isError') else False
            }
        except Exception as e:
            logger.error(f"调用工具 {tool_name} 失败: {e}")
            raise
    
    async def get_prompt(self, prompt_name: str, arguments: Optional[Dict] = None) -> Dict:
        """获取MCP prompt"""
        session = await self.ensure_connected()
        try:
            result = await session.get_prompt(prompt_name, arguments or {})
            return {
                "messages": result.messages if hasattr(result, 'messages') else [],
                "description": result.description if hasattr(result, 'description') else ""
            }
        except Exception as e:
            logger.error(f"获取prompt {prompt_name} 失败: {e}")
            raise
    
    async def close(self):
        """关闭连接"""
        try:
            if self._session:
                await self._session.__aexit__(None, None, None)
                self._session = None
            
            if self._stdio_context:
                await self._stdio_context.__aexit__(None, None, None)
                self._stdio_context = None
            
            self._read = None
            self._write = None
            self._initialized = False
            logger.info("MCP连接已关闭")
        except Exception as e:
            logger.error(f"关闭MCP连接时出错: {e}")


class MCPExecutor(Executor):
    """MCP (Model Context Protocol) executor
    
    Supports communication with local MCP servers via stdio using the official MCP SDK.
    Maintains its own LLM instance for interacting with MCP server and supports
    custom prompts for better business results.
    """
    
    def __init__(self, config: Dict[Text, Any]) -> None:
        super().__init__(config)
        
        if not MCP_AVAILABLE:
            raise ImportError(
                "MCP SDK is required. Please install it with: pip install mcp"
            )
        
        # Initialize LLM instance from config
        llm_config = self.config.get("llm", {})
        if not llm_config:
            raise ValueError("LLM configuration is required for MCP executor")
        
        self.llm = LLM(llm_config)
        logger.info(f"Initialized LLM for MCP executor: {llm_config.get('model', 'unknown')}")
        
        # MCP server configuration
        self.server_command = self.config.get("server_command", [])
        if not self.server_command:
            raise ValueError("MCP server_command is required (list of command and args)")
        
        self.custom_prompt = self.config.get("prompt", None)
        self.max_tokens = self.config.get("max_tokens", 500)
        
        # Create MCP client wrapper
        server_params = StdioServerParameters(
            command=self.server_command[0],
            args=self.server_command[1:] if len(self.server_command) > 1 else []
        )
        self._mcp_client: Optional[MCPClientWrapper] = MCPClientWrapper(server_params)
    
    async def execute(self, data: Dict[Text, Any]) -> Tuple[Text, Dict]:
        """Execute MCP operation
        
        The executor uses LLM to determine how to interact with MCP server based on:
        1. Available tools/prompts from MCP server
        2. Custom prompt (if configured)
        3. Input data
        
        Args:
            data: Data required for execution (form slots, etc.)
            
        Returns:
            Tuple[Response text, Metadata]
        """
        try:
            # Step 1: Get available tools and prompts from MCP server
            tools = await self._mcp_client.list_tools()
            prompts = await self._mcp_client.list_prompts()
            
            # Step 2: Analyze if tools/prompts are needed using LLM
            needs_tools = await self._analyze_need_for_tools(data, tools, prompts)
            
            if not needs_tools.get("needs_tools", False):
                # If no tools needed, use LLM directly
                direct_response = await self._get_direct_llm_response(data)
                return direct_response, {"mcp_action": "direct_response", "reasoning": needs_tools.get("reasoning", "")}
            
            # Step 3: Use LLM to select appropriate tool/prompt and generate arguments
            action = await self._select_tool_or_prompt(data, tools, prompts, needs_tools)
            
            # Step 4: Execute action on MCP server
            result_text = ""
            metadata = {
                "mcp_action": action.get("type"),
                "selection_reasoning": action.get("reasoning", "")
            }
            
            try:
                if action.get("type") == "call_tool":
                    tool_name = action.get("tool_name")
                    tool_args = action.get("arguments", {})
                    tool_result = await self._mcp_client.call_tool(tool_name, tool_args)
                    
                    # Extract text from content array
                    content_items = tool_result.get("content", [])
                    if content_items:
                        text_parts = []
                        for item in content_items:
                            if isinstance(item, dict):
                                text_parts.append(item.get("text", str(item)))
                            else:
                                text_parts.append(str(item))
                        result_text = "\n".join(text_parts)
                    else:
                        result_text = str(tool_result)
                    
                    metadata["tool_result"] = tool_result
                    
                    # Step 5: Use LLM to integrate tool result and generate final response
                    final_response = await self._generate_final_response(data, tool_result, result_text)
                    return final_response, metadata
                    
                elif action.get("type") == "get_prompt":
                    prompt_name = action.get("prompt_name")
                    prompt_args = action.get("arguments", {})
                    prompt_result = await self._mcp_client.get_prompt(prompt_name, prompt_args)
                    
                    # Extract text from messages
                    messages = prompt_result.get("messages", [])
                    if messages:
                        text_parts = []
                        for msg in messages:
                            if isinstance(msg, dict):
                                text_parts.append(msg.get("content", {}).get("text", str(msg)))
                            else:
                                text_parts.append(str(msg))
                        result_text = "\n".join(text_parts)
                    else:
                        result_text = prompt_result.get("description", str(prompt_result))
                    
                    metadata["prompt_result"] = prompt_result
                    return result_text, metadata
                    
                else:
                    # Fallback to direct LLM response
                    direct_response = await self._get_direct_llm_response(data)
                    return direct_response, metadata
                    
            except Exception as tool_error:
                # If tool execution fails, fallback to direct LLM response
                logger.warning(f"Tool execution failed, falling back to direct LLM: {tool_error}")
                direct_response = await self._get_direct_llm_response(data)
                metadata["tool_error"] = str(tool_error)
                return direct_response, metadata
            
        except Exception as e:
            error_msg = f"Failed to execute MCP operation: {str(e)}"
            logger.error(error_msg, exc_info=True)
            # Final fallback: try to get direct LLM response
            try:
                direct_response = await self._get_direct_llm_response(data)
                return direct_response, {"error": error_msg, "fallback": True}
            except:
                return "", {"error": error_msg}
    
    async def _analyze_need_for_tools(self, data: Dict, tools: List[Dict], prompts: List[Dict]) -> Dict:
        """使用LLM分析是否需要调用工具"""
        analysis_prompt = f"""分析以下用户输入，判断是否需要调用外部工具或服务。

用户输入数据:
{json.dumps(data, indent=2, ensure_ascii=False)}

可用工具列表: {[tool.get('name') for tool in tools]}
可用Prompts列表: {[prompt.get('name') for prompt in prompts]}

请分析：
1. 这个输入是否需要实时数据、计算或外部服务？
2. 如果需要，应该调用什么类型的工具或prompt？
3. 请用JSON格式返回分析结果，包含以下字段：
   - needs_tools: boolean (是否需要工具)
   - tool_type: string (工具类型，如calculator, weather, database等)
   - reasoning: string (分析理由)

如果不需要工具，返回：
{{"needs_tools": false, "tool_type": "", "reasoning": "不需要工具的原因"}}
"""
        
        try:
            llm_response = await self.llm.generate_chat(
                messages=[{"role": "user", "content": analysis_prompt}],
                max_tokens=300,
                response_format={"type": "json_object"}
            )
            
            analysis = json.loads(llm_response.get("content", "{}"))
            return analysis
        except Exception as e:
            logger.warning(f"分析工具需求失败，默认认为需要工具: {e}")
            return {"needs_tools": True, "tool_type": "unknown", "reasoning": f"分析失败: {str(e)}"}
    
    async def _select_tool_or_prompt(self, data: Dict, tools: List[Dict], prompts: List[Dict], analysis: Dict) -> Dict:
        """使用LLM选择合适的工具和参数"""
        tools_info = "\n".join([
            f"- {tool.get('name')}: {tool.get('description', '')}"
            for tool in tools
        ])
        
        prompts_info = "\n".join([
            f"- {prompt.get('name')}: {prompt.get('description', '')}"
            for prompt in prompts
        ])
        
        selection_prompt = f"""基于以下信息，选择合适的工具或prompt并生成调用参数。

用户输入数据:
{json.dumps(data, indent=2, ensure_ascii=False)}

分析结果: {json.dumps(analysis, indent=2, ensure_ascii=False)}

可用工具:
{tools_info}

可用Prompts:
{prompts_info}

请返回JSON格式，包含：
{{
    "type": "call_tool" 或 "get_prompt",
    "tool_name": "选中的工具名称" 或 "prompt_name": "选中的prompt名称",
    "arguments": {{"参数名": "参数值"}},
    "reasoning": "选择理由"
}}

只返回JSON，不要其他内容。"""
        
        if self.custom_prompt:
            system_prompt = self.custom_prompt
        else:
            system_prompt = "你是一个智能助手，能够分析用户需求并选择合适的工具来完成任务。"
        
        try:
            llm_response = await self.llm.generate_chat(
                messages=[
                    {"role": "system", "content": system_prompt},
                    {"role": "user", "content": selection_prompt}
                ],
                max_tokens=self.max_tokens,
                response_format={"type": "json_object"}
            )
            
            return self._parse_llm_response(llm_response.get("content", ""))
        except Exception as e:
            logger.error(f"选择工具失败: {e}")
            return {"type": "direct_response", "reasoning": f"选择失败: {str(e)}"}
    
    async def _generate_final_response(self, original_data: Dict, tool_result: Dict, tool_text: str) -> str:
        """使用LLM整合工具结果生成最终回答"""
        final_prompt = f"""基于工具调用结果，给用户一个完整、自然的回答。

用户原始输入:
{json.dumps(original_data, indent=2, ensure_ascii=False)}

工具调用结果:
{json.dumps(tool_result, indent=2, ensure_ascii=False)}

工具返回文本:
{tool_text}

请生成一个自然、完整的回答，直接回答用户的问题，不要重复说明调用了什么工具。"""
        
        try:
            llm_response = await self.llm.generate_chat(
                messages=[{"role": "user", "content": final_prompt}],
                max_tokens=self.max_tokens
            )
            return llm_response.get("content", tool_text)
        except Exception as e:
            logger.warning(f"生成最终回答失败，返回工具结果: {e}")
            return tool_text
    
    async def _get_direct_llm_response(self, data: Dict) -> str:
        """获取直接的LLM回答（不调用工具）"""
        user_input = json.dumps(data, indent=2, ensure_ascii=False)
        
        if self.custom_prompt:
            system_prompt = self.custom_prompt
        else:
            system_prompt = "你是一个有用的助手，请根据用户输入提供准确、有帮助的回答。"
        
        try:
            llm_response = await self.llm.generate_chat(
                messages=[
                    {"role": "system", "content": system_prompt},
                    {"role": "user", "content": f"用户输入:\n{user_input}\n\n请提供回答。"}
                ],
                max_tokens=self.max_tokens
            )
            return llm_response.get("content", "")
        except Exception as e:
            logger.error(f"直接LLM回答失败: {e}")
            return ""
    
    def _parse_llm_response(self, content: str) -> Dict:
        """Parse LLM response to extract action"""
        try:
            # Try to parse as JSON
            response = json.loads(content)
            return response
        except json.JSONDecodeError:
            # Try to extract JSON from string
            try:
                import re
                json_match = re.search(r'\{[^{}]*\}', content, re.DOTALL)
                if json_match:
                    response = json.loads(json_match.group())
                    return response
            except:
                pass
            
            # If all else fails, return direct response
            logger.warning(f"Failed to parse LLM response as JSON, using as direct response: {content}")
            return {
                "type": "direct_response",
                "content": content
            }
    
    async def cleanup(self) -> None:
        """Clean up MCP server resources"""
        if self._mcp_client:
            await self._mcp_client.close()
            self._mcp_client = None
        logger.info("MCP executor cleaned up")




import os
import sys
import re
import json
import logging
from pathlib import Path

from typing import Optional, Text, Dict
from cota.store import Store
from cota.agent import Agent
from cota.llm import LLM
from cota.message.message import Message
from cota.utils.io import read_yaml_from_path

logger = logging.getLogger(__name__)

class Task:
    def __init__(
            self,
            description: Optional[Text] = None,
            prompt: Optional[Text] = None,
            agents: Optional[Dict] = None,
            llm: Optional[LLM] = None
    ) -> None:
        self.description = description
        self.prompt = prompt
        self.agents = agents
        self.llm = llm  # 直接存储LLM实例

    @classmethod
    def load_from_path(cls, path:Text) -> 'Task':
        logger.debug(f"Loading task config from path: {path}")

        # load task config
        task_config = read_yaml_from_path(os.path.join(path, 'task.yml'))
        endpoints_config = read_yaml_from_path(os.path.join(path, 'endpoints.yml'))

        description = task_config.get("description")
        prompt = task_config.get("prompt")
        llm_name = task_config.get("llm")

        llm = None
        if llm_name:
            llms_config = endpoints_config.get('llms', {})
            llm_config = llms_config.get(llm_name)
            if llm_config is None:
                raise ValueError(f"LLM '{llm_name}' specified in task.yml not found in endpoints.yml")
            logger.debug(f"Initializing task LLM: {llm_name}")
            llm = LLM(llm_config)
        else:
            logger.warning("No LLM specified in task.yml")
            
        store = Store.create(endpoints_config.get('base_store', {}))
        agents = cls.load_agents(path, store)
        logger.debug(f"Task Config: \n {task_config}")

        return cls(
            description = description,
            prompt = prompt,
            agents = agents,
            llm = llm
        )

    @classmethod
    def load_agents(cls, path, store: Optional[Store]=None):
        agents = {}
        agents_path = os.path.join(path,'agents')
        for item in os.listdir(agents_path):
            agent_path = os.path.join(agents_path, item)
            agent = Agent.load_from_path(agent_path, store)
            agents[agent.name] = agent
        return agents

    def get_llm(self) -> LLM:
        if self.llm is None:
            raise ValueError("No LLM configured for this task. Please specify 'llm' field in task.yml")
        return self.llm

    async def run(self):
        if self.prompt:
            logger.debug(f"Generating tasks through LLM...")
            await self.run_with_llm()
        else:
            logger.error("No prompt provided for task generation.")
            sys.exit(1)


    async def run_with_llm(self):
        # Generate plan
        next_plan = await self.generate_plans()
        while True:
            await self.execute_task(next_plan)
            next_plan = await self.generate_plans()

    async def execute_task(self, task):
        logger.debug(f"Executing task {task}")
        agent = self.agents.get(task.get('agent'))
        await agent.processor.handle_session('test_001')



    async def generate_plans(self) -> Dict:
        logger.debug(f"Generating a DAG plans through LLM...")
        prompt = self.format_prompt(self.prompt)

        llm = self.get_llm()
        result = await llm.generate_chat(
            messages = [{"role": "system", "content": "You are a task planner, good at breaking down tasks into DAG execution flows"},{"role":"user", "content": prompt}],
            max_tokens = 1000,
            response_format = {'type': 'json_object'}
        )
        print('result: ', result)
        plans = json.loads(result["content"])

        logger.debug(f"Generating plans prompt: {prompt}")

        return plans

    def format_prompt(self, prompt: Text) -> Text:
        def observe(name):
            if hasattr(self, name):
                method = getattr(self, name)
                return method()
            else:
                raise AttributeError(f"Method {name} not found")

        variable_names = re.findall(r'\{\{(\w+)\}\}', prompt)
        format_dict = {var_name: observe(var_name) for var_name in variable_names}
        for key in format_dict:
            prompt = prompt.replace('{{' + key + '}}', format_dict[key])
        return prompt

    def agent_description(self) -> Text:
        description = ""
        for _, agent in self.agents.items():
            description = description + agent.name + ":" + agent.description + '\n'
        description = description + '\n'
        return description

    def task_description(self) -> Text:
        return self.description
    
    def history_messages(self) -> Text:
        merged_messages = set()
        for name, agent in self.agents.items():
            if agent.processor.dst:
                state = agent.processor.dst.current_state()
                logger.debug(f"Task DST State {state}")
                for action in state.get('actions'):
                    for result in action.get('result'):
                        merged_messages.add(
                            (action.get('timestamp'),result.get('sender_id') + ':' + result.get('text',''))
                        )
        messages = sorted(list(merged_messages),  key=lambda x: x[0])
        result = '\n'.join([message[1] for message in messages])
        logger.debug(f"History Message {result}")
        return result
    


import asyncio
import logging
import time
from typing import Any, Dict, List, Optional, Set

logger = logging.getLogger(__name__)

class RoomManager:

    def __init__(self):
        self._rooms: Dict[str, Dict[str, Any]] = {}
        self._user_rooms: Dict[str, Set[str]] = {}
        self._lock = asyncio.Lock()
    
    async def create_or_update_room(self, room_id: str, user_ids: List[str]) -> bool:
        room_id = str(room_id)
        user_ids = [str(uid) for uid in user_ids]
        
        async with self._lock:
            self._rooms[room_id] = {
                "user_ids": user_ids,
                "created_at": self._rooms.get(room_id, {}).get("created_at", time.time()),
                "updated_at": time.time(),
                "last_activity": time.time()
            }
            
            # Update user-room map
            for user_id in user_ids:
                if user_id not in self._user_rooms:
                    self._user_rooms[user_id] = set()
                self._user_rooms[user_id].add(room_id)
            
            # Iterate through all users that currently have room associations
            # We use list() to create a static copy for safe iteration while modifying
            for user_id in list(self._user_rooms.keys()):
                # Check if the user is not in the updated room member list but still has this room association
                if user_id not in user_ids and room_id in self._user_rooms[user_id]:
                    # Remove this room from the user's room set
                    self._user_rooms[user_id].remove(room_id)
                    
                    # If the user has no remaining room associations after removal
                    if not self._user_rooms[user_id]:
                        # Delete the user entry entirely to prevent empty sets and maintain memory efficiency
                        del self._user_rooms[user_id]
            
            logger.debug(f"Updated room {room_id} with users: {user_ids}")
            return True
    
    async def get_room(self, room_id: str) -> Optional[Dict[str, Any]]:
        room_id = str(room_id)
        async with self._lock:
            room = self._rooms.get(room_id)
            if room:
                room["last_activity"] = time.time()
            return room
    
    async def get_room_users(self, room_id: str) -> List[str]:
        room_id = str(room_id)
        async with self._lock:
            room = self._rooms.get(room_id)
            return room["user_ids"] if room else []
    
    async def get_user_rooms(self, user_id: str) -> List[str]:
        user_id = str(user_id)
        async with self._lock:
            return list(self._user_rooms.get(user_id, set()))
    
    async def remove_room(self, room_id: str) -> bool:
        room_id = str(room_id)
        async with self._lock:
            if room_id in self._rooms:
                user_ids = self._rooms[room_id]["user_ids"]
                del self._rooms[room_id]
                
                # Update user-room map
                for user_id in user_ids:
                    if user_id in self._user_rooms:
                        self._user_rooms[user_id].discard(room_id)
                        if not self._user_rooms[user_id]:
                            del self._user_rooms[user_id]
                
                logger.debug(f"Removed room {room_id}")
                return True
            return False
    
    async def cleanup_inactive(self, timeout: int = 3600) -> int:
        now = time.time()
        to_remove = []
        
        async with self._lock:
            for room_id, room in self._rooms.items():
                if now - room["last_activity"] > timeout:
                    to_remove.append(room_id)
        
        count = 0
        for room_id in to_remove:
            if await self.remove_room(room_id):
                count += 1
        
        return count

    async def add_user_to_room(self, room_id: str, user_id: str) -> None:
        """Add a user to a room"""
        room = await self.get_or_create_room(room_id)
        if user_id not in room["user_ids"]:
            room["user_ids"].append(user_id)
            room["updated_at"] = time.time()
            room["last_activity"] = time.time()
            
            # Update user-room map
            if user_id not in self._user_rooms:
                self._user_rooms[user_id] = set()
            self._user_rooms[user_id].add(room_id)
            
            # Iterate through all users that currently have room associations
            # We use list() to create a static copy for safe iteration while modifying
            for user_id in list(self._user_rooms.keys()):
                # Check if the user is not in the updated room member list but still has this room association
                if user_id not in room["user_ids"] and room_id in self._user_rooms[user_id]:
                    # Remove this room from the user's room set
                    self._user_rooms[user_id].remove(room_id)
                    
                    # If the user has no remaining room associations after removal
                    if not self._user_rooms[user_id]:
                        # Delete the user entry entirely to prevent empty sets and maintain memory efficiency
                        del self._user_rooms[user_id]
            
            logger.debug(f"Updated room {room_id} with users: {room['user_ids']}")
    
    async def remove_user_from_room(self, room_id: str, user_id: str) -> None:
        """Remove a user from a room"""
        room = await self.get_room(room_id)
        if room and user_id in room["user_ids"]:
            room["user_ids"].remove(user_id)
            room["updated_at"] = time.time()
            room["last_activity"] = time.time()
            
            # If the room is empty, delete the room
            if not room["user_ids"]:
                await self.remove_room(room_id)
            
            # Update user-room map
            if user_id in self._user_rooms:
                self._user_rooms[user_id].discard(room_id)
                if not self._user_rooms[user_id]:
                    del self._user_rooms[user_id]
            
            logger.debug(f"Updated room {room_id} with users: {room['user_ids']}")
    
    async def get_or_create_room(self, room_id: str) -> Dict[str, Any]:
        """Get or create a room"""
        room = await self.get_room(room_id)
        if room is None:
            room = {
                "user_ids": [],
                "created_at": time.time(),
                "updated_at": time.time(),
                "last_activity": time.time()
            }
            self._rooms[room_id] = room
        return room


import json
import os
import copy
import logging
from typing import Text, List, Union, Optional, Dict, Tuple
from cota.actions.action import Action
from cota.actions.form import Form
from cota.actions.user_utter import UserUtter
from cota.actions.executors.base import Executor
from cota.dst import DST
from cota.utils.io import read_yaml_from_path
from cota.processor import Processor
from cota.store import Store, MemoryStore, SQLStore
from cota.llm import LLM
from cota.dpl.dpl import DPL, DPLFactory
from cota.knowledge.knowledge import KnowledgeFactory, Knowledge
from cota.utils.http import HttpClientManager, HttpConfig
from cota.utils.common import (
    first_empty_key,
    merge_dicts,
    hash_str
)

from cota.constant import (
    DEFAULT_CONFIG,
    DEFAULT_FORM_CONFIG,
    DEFAULT_HTTP_CLIENT_CONFIG
)

logger = logging.getLogger(__name__)

class Agent:
    def __init__(
            self,
            name: Optional[Text] = None,
            description: Optional[Text] = None,
            actions: Optional[Dict] = None,
            llms: Optional[Dict[Text, LLM]] = None,
            dpl: Optional[DPL] = None,
            store: Optional[Store] = None,
            dialogue: Optional[Dict] = None,
            user_proxy: Optional[Dict] = None,
            knowledge: Optional[Knowledge] = None
    ) -> None:
        self.name = name
        self.description = description
        self.actions = actions
        self.llms = llms
        self.dpl = dpl
        self.store = store
        self.dialogue = dialogue
        self.user_proxy = user_proxy
        self.knowledge = knowledge
        self.processor = Processor(agent=self, store=self.store)
        self._executors = {}  # Dictionary to store executor instances

    @classmethod
    def load_from_path(cls, path: Text, store: Optional[Store] = None) -> "Agent":
        """
        Load agent configuration from the specified path and create an Agent instance.
        
        Args:
            path (Text): Path to the agent configuration directory
            store (Optional[Store]): Optional store instance, will create one if not provided
            
        Returns:
            Agent: Configured agent instance
        """
        logger.debug(f"Loading agent config from path: {path}")

        # Load and merge configuration files
        agent_config = read_yaml_from_path(os.path.join(path, 'agent.yml'))
        endpoints_config = read_yaml_from_path(os.path.join(path, 'endpoints.yml'))
        agent_config = cls.merge_agent_config(DEFAULT_CONFIG, agent_config)

        # Extract core configuration
        system_config = agent_config.get("system", {})
        name = system_config.get("name")
        description = system_config.get("description") 
        actions = agent_config.get("actions", {})
        dialogue = agent_config.get("dialogue", {})
        user_proxy = agent_config.get("user_proxy", {})

        # Initialize or use provided store
        if not store:
            store_config = endpoints_config.get('base_store', {})
            store = Store.create(store_config)

        # Initialize language models
        llms = {}
        for llm_name, config in endpoints_config.get('llms', {}).items():
            logger.debug(f"Initializing language model: {llm_name}")
            llms[llm_name] = LLM(config)

        # Initialize dialogue policy learning
        policy_path = os.path.join(path, "policy")
        dpl = DPLFactory.create(agent_config, policy_path)

        # Initialize knowledge list
        knowledge_configs = agent_config.get('knowledge', [])
        knowledge = KnowledgeFactory.create(knowledge_configs, path)

        # Initialize executors
        executors = cls._init_executors(actions)

        # Log final configurations for debugging
        logger.debug(f"Endpoints configuration:\n{endpoints_config}")
        logger.debug(f"Agent configuration:\n{agent_config}")

        agent = cls(
            name=name,
            description=description, 
            actions=actions,
            store=store,
            llms=llms,
            dpl=dpl,
            dialogue=dialogue,
            user_proxy=user_proxy,
            knowledge=knowledge
        )
        
        agent._executors = executors
        
        return agent

    @classmethod
    def _init_executors(cls, actions: Dict) -> Dict[Text, Executor]:
        """Initialize executors
        
        Args:
            actions: Action configuration dictionary
            
        Returns:
            Dict[Text, Executor]: Executor dictionary
        """
        executors = {}
        
        for action_name, action_config in actions.items():
            if "executer" in action_config:
                executor_config = action_config["executer"]
                executor_type = executor_config.get("type", "http")  # Default to http type
                
                try:
                    executor = Executor.create(executor_type, executor_config)
                    executors[action_name] = executor
                except Exception as e:
                    logger.error(f"Failed to initialize executor for {action_name}: {str(e)}")
                    
        return executors

    def get_executor(self, action_name: Text) -> Optional[Executor]:
        """Get executor for specified action
        
        Args:
            action_name: Action name
            
        Returns:
            Optional[Executor]: Executor instance, returns None if not exists
        """
        return self._executors.get(action_name)

    async def cleanup(self):
        """Clean up resources"""
        for executor in self._executors.values():
            if hasattr(executor, 'cleanup'):
                await executor.cleanup()

    def build_action(self, action_name: Text, **kwargs) -> "Action":
        """
        Builds an action based on the provided action name and optional parameters.

        Args:
            action_name (Text): The name of the action to build.
            **kwargs: Additional parameters specific to each action type.

        Returns:
            Action: The built action instance.

        Raises:
            ValueError: If the action is not found in the configuration.
        """
        # Try exact match first, then case-insensitive match
        action_config = self.actions.get(action_name)
        if action_config is None:
            # Try case-insensitive match
            action_name_lower = action_name.lower()
            for key, config in self.actions.items():
                if key.lower() == action_name_lower:
                    action_config = config
                    action_name = key  # Use the actual key name from config
                    break
        if action_config is None:
            raise ValueError(f"Action '{action_name}' not found in actions configuration.")

        # Build parameters from config and kwargs
        params = {
            "name": action_name,
            "description": kwargs.get("description") or action_config.get("description") or action_name,
            "prompt": kwargs.get("prompt") or action_config.get("prompt"),
            "llm": kwargs.get("llm") or action_config.get("llm"),
            "sender_id": self.name
        }

        # Add form-specific parameters if needed
        if action_config.get("type") == "form" or "executer" in action_config:
            slots = {key: '' for key in action_config.get("slots", {})}
            params.update({
                "slots": kwargs.get("slots") or {key: '' for key in action_config.get("slots", {})},
                "state": kwargs.get("state"),
                "executer": action_config.get("executer")
            })
            return Form.build_from_name(**params)

        # Add any additional kwargs that might be specific to other action types
        params.update(kwargs)
        return Action.build_from_name(**params)

    async def generate_actions(self, dst: DST) -> List[Action]:
        """Generate the corresponding action based on DPL"""
        if dst.current_form:
            return await self._handle_current_form(dst)

        # First try to generate action using DPL action generator
        if self.dpl:
            action_names = await self.dpl.generate_actions(dst)
            if action_names:
                # Only take the first action as we only support single action
                first_action_name = action_names[0]
                return [self.build_action(first_action_name)]

        # If DPL doesn't generate action, fall back to selector
        selector = self.build_action(
            action_name='Selector'
        )
        await selector.run(agent=self, dst=dst)
        dst.update(selector)

        if len(selector.result) == 0:
            # if no action is selected, return a Response action
            return [self.build_action('BotUtter')]
        else:
            # if actions are selected, take only the first action
            action_infos = self._extract_action_info(selector)
            if action_infos:
                action_name, action_params = action_infos[0]
                return [self.build_action(action_name, **action_params)]
            else:
                return [self.build_action('BotUtter')]

    async def _handle_form_query(self, dst: DST) -> List[Action]:
        selector = self.build_action(
            action_name='Selector'
        )
        await selector.run(agent = self, dst=dst)
        dst.update(selector)
        if len(selector.result) == 0:
            return [self.build_action('BotUtter')]

        action_infos = self._extract_action_info(selector)
        if action_infos:
            action_name, action_params = action_infos[0]
            if dst.current_form.name == action_name:
                return [self.build_action(
                    action_name=dst.current_form.name,
                    slots=action_params.get('slots', {}),
                    state="continue"
                )]
            else:
                # break current form and execute the new action
                return [
                    self.build_action(
                        action_name=dst.current_form.name,
                        slots=copy.deepcopy(dst.current_form.slots),
                        state="break"
                    ),
                    self.build_action(
                        action_name=action_name
                    )
                ]
        else:
            return [self.build_action('BotUtter')]
    
    async def _handle_current_form(self, dst: DST) -> List[Action]:
        if isinstance(dst.latest_action, UserUtter):
            return await self._handle_form_query(dst)

        required_slot = first_empty_key(dst.current_form.slots)
        if not required_slot:
            return [
                self.build_action(
                    action_name=dst.current_form.name,
                    slots=copy.deepcopy(dst.current_form.slots),
                    state="continue"
                )]

        action_config = self.actions.get(dst.current_form.name, {})
        slots_config = action_config.get("slots", {})
        return [ 
                self.build_action(
                    action_name='BotUtter',
                    description = slots_config.get(required_slot,{}).get("description",{}),
                    prompt = slots_config.get(required_slot,{}).get("prompt",{}),
                    llm = slots_config.get(required_slot,{}).get("llm",None)
                )]

    def _extract_action_info(self, selector: Action) -> List[Dict]:
        action_info = []
        
        for d in selector.result:
            action_name = d.get('text')
            metadata = d.get('metadata', {})
            
            # Extract only the fields that should be passed as action parameters
            # Based on selector implementation, result now contains 'text', 'thought' and 'slots'
            # Only 'slots' should be passed as action parameters
            action_params = {}
            if 'slots' in metadata and metadata['slots']:
                action_params['slots'] = metadata['slots']
            
            logger.debug(f"Action: {action_name}, Original metadata: {metadata}")
            logger.debug(f"Action: {action_name}, Extracted action params: {action_params}")
            
            action_info.append((action_name, action_params))
        return action_info


    def llm_instance(self, llm_name:Optional[Text] = None) -> LLM:
        if llm_name is None:
            first_llm_name, first_llm_config = next(iter(self.llms.items()))
            return first_llm_config

        llm_config = self.llms.get(llm_name)
        if llm_config is None:
            raise ValueError(f"LLM with name '{llm_name}' not found.")
        return llm_config

    @classmethod
    def merge_agent_config(cls, default_config, config):
        """
        Merges the provided config with the default config.
        """
        import copy
        actions = config.get('actions', {})
        for action_name in actions:
            if 'executer' in actions[action_name]: 
                default_form_config = copy.deepcopy(DEFAULT_FORM_CONFIG)
                actions[action_name] = merge_dicts(default_form_config, actions[action_name])
        default_config = copy.deepcopy(DEFAULT_CONFIG)
        updated_dict = merge_dicts(default_config, config)
        return updated_dict

    def create_processor(self) -> Processor:
        """
        Creates and returns a new Processor instance.
        
        Returns:
            Processor: A new processor instance initialized with this agent and its store.
        """
        processor = Processor(
            agent=self,
            store=self.store
        )
        return processor


import logging
import re
import json
from typing import Text, Optional, List, Dict, Any
from cota.actions.action import Action
from cota.dst import DST
from cota.utils.parser import extract_action_from_string
from cota.constant import DEFAULT_SELECTOR_INSTRUCTION
from cota.constant import (
    DEFAULT_DIALOGUE_MAX_TOKENS
)

logger = logging.getLogger(__name__)

class Selector(Action):
    """Selector class for selecting the next action to execute
    """
    def __init__(
            self,
            name: Optional[Text] = None,
            description: Optional[Text] = None,
            prompt: Optional[Text] = None,
            llm: Optional[Text] = None,
            timestamp: Optional[float] = None,
            sender_id: Optional[Text] = None,
            receiver_id: Optional[Text] = None,
            result: Optional[List] = None,
            **kwargs
    ) -> None:
        """Initialize the selector
        
        Args:
            name: Name of the selector
            description: Description of the selector
            prompt: Prompt text
            llm: Language model to use
            timestamp: Timestamp
            sender_id: Sender ID
            receiver_id: Receiver ID
            result: Result list
        """
        super().__init__(
            name=name,
            description=description,
            prompt=prompt,
            llm=llm,
            timestamp=timestamp,
            sender_id=sender_id,
            receiver_id=receiver_id,
            result=result
        )

    def apply_to(self, dst: DST) -> None:
        dst.actions.append(self)
        #dst.formless_actions.append(self)
        dst.latest_action = self

    def run_from_string(self, text: Text) -> None:
        self.result.append(
            {
                "action_name": text,
            }
        )


    async def run(
            self,
            agent: Optional["Agent"] = None,
            dst: Optional[DST] = None
    ):
        # System prompt
        system_prompt = DEFAULT_SELECTOR_INSTRUCTION
        # Enhanced prompt
        knowledge_dict = await dst.format_knowledge(self.prompt, self)
        thoughts_dict = await dst.format_policies(self.prompt, self)
        prompt = dst.format_prompt(self.prompt, self, {**knowledge_dict, **thoughts_dict})

        select_result = await agent.llm_instance(self.llm).generate_chat(
            messages=[
                {"role": "system", "content": DEFAULT_SELECTOR_INSTRUCTION},
                {"role": "user", "content": prompt}
            ],
            max_tokens=agent.dialogue.get('max_tokens', DEFAULT_DIALOGUE_MAX_TOKENS),
            response_format = {'type': 'json_object'}
        )
        # Extract content from result
        content = select_result["content"]
        try:
            select_result = json.loads(content)
        except json.JSONDecodeError:
            logger.error(f"Failed to parse JSON response from selector: {content}")

        logger.debug(f"Selector result before: {select_result}")
        
        result_dict = {"text": select_result.get('action')}
        if select_result.get('thought'):
            result_dict['thought'] = select_result.get('thought')
        
        self.result = [result_dict]

        logger.debug(f"Selector prompt: {prompt}")
        logger.debug(f"Selector result: {self.result}")


import logging
import asyncio
import aiohttp
from typing import Optional, Dict, Any, Union
from dataclasses import dataclass
from aiohttp import ClientTimeout

logger = logging.getLogger(__name__)

@dataclass
class HttpConfig:
    """HTTP client configuration"""
    max_retries: int = 3
    timeout: int = 300  # 5 minutes
    base_url: str = ""
    default_headers: Dict[str, str] = None
    
    def __post_init__(self):
        if self.default_headers is None:
            self.default_headers = {"Content-Type": "application/json"}

class HttpResponse:
    """HTTP response wrapper"""
    def __init__(self, status: int, data: Any, headers: Dict[str, str]):
        self.status = status
        self.data = data
        self.headers = headers
        
    @property
    def ok(self) -> bool:
        return 200 <= self.status < 300

class HttpClient:
    """HTTP client implementation
    
    Provides configurable HTTP client implementation, supports:
    - Automatic retry
    - Timeout control
    - Unified error handling
    - Request/response interception
    """
    
    def __init__(self, config: Optional[HttpConfig] = None):
        self.config = config or HttpConfig()
        self._session: Optional[aiohttp.ClientSession] = None
        
    async def __aenter__(self):
        await self.connect()
        return self
        
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()
        
    async def connect(self):
        """Create connection session"""
        if not self._session:
            timeout = ClientTimeout(total=self.config.timeout)
            self._session = aiohttp.ClientSession(timeout=timeout)
            
    async def close(self):
        """Close connection session"""
        if self._session:
            await self._session.close()
            self._session = None
    
    def _build_url(self, path: str) -> str:
        """Build complete request URL"""
        if path.startswith(("http://", "https://")):
            return path
        return f"{self.config.base_url.rstrip('/')}/{path.lstrip('/')}"
    
    async def request(
        self,
        method: str,
        url: str,
        *,
        params: Optional[Dict] = None,
        json_data: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        **kwargs
    ) -> HttpResponse:
        """Send HTTP request
        
        Args:
            method: HTTP method
            url: Request path or complete URL
            params: URL query parameters
            json_data: JSON request body
            headers: Request headers
            **kwargs: Other parameters supported by aiohttp
            
        Returns:
            HttpResponse object
            
        Raises:
            Exception: Thrown when still fails after retries
        """
        if not self._session:
            await self.connect()
            
        merged_headers = {**self.config.default_headers, **(headers or {})}
        full_url = self._build_url(url)
        last_error = None
        
        for attempt in range(self.config.max_retries):
            try:
                async with self._session.request(
                    method,
                    full_url,
                    params=params,
                    json=json_data,
                    headers=merged_headers,
                    **kwargs
                ) as response:
                    data = await response.json()
                    return HttpResponse(
                        status=response.status,
                        data=data,
                        headers=dict(response.headers)
                    )
            except Exception as e:
                last_error = e
                logger.warning(
                    f"HTTP request attempt {attempt + 1} failed: {str(e)}",
                    extra={
                        "url": full_url,
                        "method": method,
                        "attempt": attempt + 1
                    }
                )
                if attempt < self.config.max_retries - 1:
                    await asyncio.sleep(1 * (attempt + 1))  # Exponential backoff
                continue
                
        raise Exception(
            f"HTTP request failed after {self.config.max_retries} attempts: {str(last_error)}"
        )
    
    async def get(self, url: str, **kwargs) -> HttpResponse:
        """Send GET request"""
        return await self.request("GET", url, **kwargs)
        
    async def post(self, url: str, **kwargs) -> HttpResponse:
        """Send POST request"""
        return await self.request("POST", url, **kwargs)
        
    async def put(self, url: str, **kwargs) -> HttpResponse:
        """Send PUT request"""
        return await self.request("PUT", url, **kwargs)
        
    async def delete(self, url: str, **kwargs) -> HttpResponse:
        """Send DELETE request"""
        return await self.request("DELETE", url, **kwargs) 


import os
import logging
from pathlib import Path
from typing import Text, List, Dict, Optional, Union, Tuple, Any
from cota.dst import DST
from cota.actions.action import Action
from cota.dpl.dpl import DPL
from cota.utils.io import read_yaml_from_path
from cota.utils.common import hash_str
from cota.constant import DEFAULT_DIALOGUE_MAX_TOKENS

logger = logging.getLogger(__name__)


class LLMDPL(DPL):
    """LLM-based Dialogue Policy Learning class that uses LLM with knowledge retrieval to generate thoughts."""
    
    def __init__(
        self,
        path: Union[Text, Path],
        llms: Union[Text, Dict, List, None] = None
    ) -> None:
        """Initialize LLMDPL with policy data."""
        # Support llms list configuration
        self.llms = llms or []
        
        # Build action to LLM mapping and find default LLM (similar to LLMKnowledge)
        self.action_llm_map = {}
        self.default_llm = None
        
        if isinstance(self.llms, list):
            self._build_llm_mappings()
        
        self.build(path)
    
    def _build_llm_mappings(self):
        """Build action to LLM mapping and find default LLM from llms list configuration."""
        for llm_config in self.llms:
            if not isinstance(llm_config, dict):
                logger.warning(f"Invalid LLM config format: {llm_config}")
                continue
                
            llm_name = llm_config.get('name')
            action = llm_config.get('action')
            
            if not llm_name:
                logger.warning(f"LLM configuration missing name: {llm_config}")
                continue
                
            if action:
                # LLM with specific action
                self.action_llm_map[action] = llm_name
            else:
                # LLM without specific action serves as default
                self.default_llm = llm_name
    
    def get_llm_for_action(self, action_name: str) -> Optional[str]:
        """Get LLM configuration for specific action type."""
        if not self.llms:
            return None
            
        # Handle list format (primary configuration method)
        if isinstance(self.llms, list):
            # Check if we have a specific LLM for this action
            if action_name in self.action_llm_map:
                return self.action_llm_map[action_name]
            
            # Use default LLM if available
            if self.default_llm:
                return self.default_llm
                
            # Fallback to first available LLM
            if self.llms:
                first_llm = self.llms[0]
                if isinstance(first_llm, dict):
                    return first_llm.get('name')
                    
            return None
            
        # Handle string format (single LLM for all actions) - backward compatibility
        elif isinstance(self.llms, str):
            return self.llms
            
        # Handle dictionary format (legacy direct action mapping) - backward compatibility
        elif isinstance(self.llms, dict):
            # First try exact match
            if action_name in self.llms:
                return self.llms[action_name]
            # Fallback to 'default' key if available
            return self.llms.get('default')
            
        else:
            logger.warning(f"Unsupported LLM config type: {type(self.llms)}")
            return None
    
    
    def build(self, path: Union[Text, Path]):
        """Prepare data for LLM with streaming processing."""
        # Load policy data from path
        policies = self.load_data(path)

        if not policies:
            logger.warning(f"No policies found in {path}")
            return {}

        # Stream process: read policies and write to file directly
        self.process_policies(policies, path)

        # Log reminder for LLM knowledge platform submission
        logger.info("=" * 60)
        logger.info("📋 LLM DPL Data Preparation Complete (Auto-Generated by Action Type)")
        logger.info("=" * 60)
        logger.info(f"📁 Source path: {path}")
        logger.info("")
        logger.info("⚠️  IMPORTANT: Please submit the following data to your LLM knowledge platform:")
        logger.info(f"  1. Upload all generated knowledge files to your knowledge base")
        logger.info("   2. Configure LLM knowledge retrieval for each action type")
        logger.info("   3. Test LLM knowledge query functionality for all action types")
        logger.info("📤 Ready-to-upload files automatically separated by action type")
        logger.info("=" * 60)

    def process_policies(self, policies: List[Dict], policies_path: Union[Text, Path]):
        """Process policies: read and write to file directly for all action types."""
        import time
        
        policies_path = Path(policies_path)
        timestamp = int(time.time())
        
        # First pass: collect all action types that have thoughts
        action_types = set()
        for policy in policies:
            actions = policy.get("actions", [])
            for action in actions:
                if 'thought' in action:
                    action_name = action.get('name', '')
                    if action_name:
                        action_types.add(action_name)
        
        if not action_types:
            logger.warning("No actions with thoughts found in policies")
            return
        
        # Create output files for each action type
        action_files = {}
        action_hash_keys = {}
        file_handles = {}
        
        try:
            # Initialize files and data structures for each action type
            for action_type in action_types:
                safe_action_name = action_type.lower().replace(' ', '_')
                output_path = policies_path / f"llm_dpl_{safe_action_name}_{timestamp}.md"
                output_path.parent.mkdir(parents=True, exist_ok=True)
                
                action_hash_keys[action_type] = set()
                file_handles[action_type] = open(output_path, 'w', encoding='utf-8')
                
                logger.info(f"Created knowledge file for {action_type}: {output_path.name}")
            
            # Process all policies
            for policy_idx, policy in enumerate(policies):
                actions = policy.get("actions", [])
                if not actions:
                    continue
                
                policy_title = policy.get("title", f"policy_{policy_idx}")
                user_utter_index = self.build_user_utter_index(actions)
                
                for i, action in enumerate(actions):
                    if 'thought' in action:
                        action_name = action.get('name', '')
                        
                        # Skip if action type not found or no file handle
                        if action_name not in file_handles:
                            continue
                        
                        target_file = file_handles[action_name]
                        target_hash_keys = action_hash_keys[action_name]
                        
                        segments = self.trace_back_to_user_utter(actions, i, user_utter_index)
                        
                        for segment in segments:
                            hash_key = '_'.join(f"{action.get('name')}:{action.get('thought', '')}" for action in segment)
                            if hash_str(hash_key) not in target_hash_keys:
                                target_hash_keys.add(hash_str(hash_key))
                                
                                content = self.segment_to_llm_content(segment)
                                if content:
                                    target_file.write(f"# {policy_title}\n")
                                    target_file.write(f"{content}\n")
                                    target_file.write(f"\n\n")
        
        finally:
            # Close all file handles
            for file_handle in file_handles.values():
                file_handle.close()
            
        logger.info(f"Generated knowledge files for {len(action_types)} action types: {list(action_types)}")


    def segment_to_llm_content(self, segment: List[Dict]) -> str:
        """Convert action segment to LLM knowledge document content."""
        content_lines = []
        
        for action in segment:
            action_name = action.get('name')
            if not action_name:
                continue
            
            # Add thought if available
            if action.get('thought'):
                content_lines.append(f"thought:{action.get('thought')}")

            # Add action result if available
            if action.get('result'):
                content_lines.append(f"{action_name}:{action.get('result')}")
        
        return "\n".join(content_lines)
    
    async def generate_thoughts(self, dst: DST, action: Action) -> Text:
        """Generate thoughts using LLM with action-specific configuration."""
        # Determine action name
        action_name = action.name if hasattr(action, 'name') else action.__class__.__name__
        
        # Get LLM for this action type
        llm_name = self.get_llm_for_action(action_name)
        if not llm_name:
            logger.warning(f"No LLM configured for action type: {action_name}")
            return ""
        
        # Build query from current context
        query_text = self.build_llm_query(dst, action)
        if not query_text.strip():
            logger.debug("Empty query text for LLM knowledge retrieval")
            return ""
        
        logger.debug(f"Using LLM '{llm_name}' for action '{action_name}'")
        
        try:
            result = await dst.agent.llm_instance(llm_name).generate_chat(
                messages=[
                    {
                        "role":"user", 
                        "content": query_text
                    }
                ],
                max_tokens=dst.agent.dialogue.get('max_tokens', DEFAULT_DIALOGUE_MAX_TOKENS)
            )
            return result["content"]
        except Exception as e:
            logger.error(f"Failed to generate thoughts for {action_name} using {llm_name}: {e}")
            return ""

    def build_llm_query(self, dst: DST, action: Action) -> Text:
        """Build query text for LLM knowledge retrieval using the longest context segment."""
        actions = dst.formless_actions
        segment_parts = []
        for act in actions:
            if act and act.result:
                # result is a list, need to search for text in the result items
                text_parts = []
                for result_item in act.result:
                    if isinstance(result_item, dict) and result_item.get('text'):
                        text_parts.append(result_item.get('text', ''))
                
                # Combine all text parts for this action into one entry
                if text_parts:
                    combined_text = ' '.join(text_parts)
                    segment_parts.append(f"{act.name}:{combined_text}")
        
        # Return the complete segment as query
        return "\n".join(segment_parts)
        
    def load_data(self, path: Union[Text, Path]) -> List[Dict]:
        """Load policy data from YAML files, similar to MatchDPL."""
        policies = []
        path_obj = Path(path)

        if not path_obj.is_dir():
            logger.warning(f"Path {path} is not a directory")
            return policies

        # Process all YAML files in directory
        for yml_file in path_obj.glob('*.yml'):
            try:
                data = read_yaml_from_path(yml_file)
                
                # Only extract data from 'policies' sections
                if isinstance(data, dict) and 'policies' in data:
                    policies.extend(data['policies'])
            except Exception as e:
                logger.error(f"Failed to load {yml_file}: {e}")
                
        return policies

    def build_user_utter_index(self, actions):
        return [i for i, action in enumerate(actions) if action.get('name') == 'UserUtter']

    def trace_back_to_user_utter(self, actions, index, user_utter_index):
        segments = []
        for q_index in reversed(user_utter_index):
            if q_index < index:
                segment = [action for action in actions[q_index:index] if action.get('name') and action.get('name') not in ('Selector','Updater')]
                segment.append(actions[index])
                if len(segment) > 1:
                    segments.append(segment)
        return segments



from cota.actions.action import Action
from cota.actions.user_utter import UserUtter
from cota.actions.bot_utter import BotUtter
from cota.actions.selector import Selector
from cota.actions.form import Form
from cota.actions.rag import RAG


"""COTA - Chain of Thought Agent Platform for Industrial-Grade Dialogue Systems

COTA can be used as a backend library in your Python applications.

Basic usage:
    ```python
    from cota import Agent, Message
    
    # Initialize agent from configuration directory
    agent = Agent.load_from_path(path="/path/to/agent/config")
    
    # Create a message
    message = Message(
        text="Hello, how can you help me?",
        sender="user",
        sender_id="user_123",
        session_id="session_456"
    )
    
    # Process the message
    processor = agent.create_processor()
    await processor.handle_message(message)
    ```
"""

__version__ = "1.1.1"  # same as pyproject.toml

# Core classes for library usage
from cota.agent import Agent
from cota.message.message import Message

# Re-export commonly used items
__all__ = [
    "__version__",
    "Agent",
    "Message",
]



import asyncio
import inspect
import json
import logging
import uuid
from asyncio import Queue, CancelledError
from sanic import Sanic, Blueprint, response
from sanic.request import Request
from sanic.response import HTTPResponse
from typing import Text, List, Dict, Any, Optional, Callable, Iterable, Awaitable, NoReturn

from cota.message.message import Message
from cota.channels.utils import convert_message_dict

try:
    from urlparse import urljoin  # pytype: disable=import-error
except ImportError:
    from urllib.parse import urljoin

logger = logging.getLogger(__name__)


def register(
    channels: List["Channel"], app: Sanic, route: Optional[Text] = None
) -> None:
    """Register a group of input channels to Sanic application and create corresponding route handlers for each channel"""
    async def handler(message: Message, channel):
        await app.ctx.agent.processor.handle_message(message, channel)

    for channel in channels:
        if route:
            p = urljoin(route, channel.url_prefix())
        else:
            p = None
        # registe bulueprint to sanic
        app.blueprint(channel.blueprint(handler), url_prefix=p)
    app.ctx.channels = channels


class Channel:
    @classmethod
    def name(cls) -> Text:
        return cls.__name__

    def url_prefix(self) -> Text:
        return self.name()

    def blueprint(
        self, on_new_message: Callable[[Message], Awaitable[Any]]
    ) -> Blueprint:
        """Defines a Sanic blueprint.

        The blueprint will be attached to a running sanic server and handle
        incoming routes it registered for."""
        raise NotImplementedError("Component listener needs to provide blueprint.")

    @classmethod
    def raise_missing_credentials_exception(cls) -> NoReturn:
        raise Exception(
            "To use the {} input channel, you need to "
            "pass a credentials file using '--credentials'. "
            "The argument should be a file path pointing to "
            "a yml file containing the {} authentication "
            "information. Details in the docs: "
            "{}/user-guide/messaging-and-voice-channels/".format(
                cls.name(), cls.name(), DOCS_BASE_URL
            )
        )


    def get_metadata(self, request: Request) -> Optional[Dict[Text, Any]]:
        """Extracts additional information from the incoming request.

         Implementing this function is not required. However, it can be used to extract
         metadata from the request. The return value is passed on to the
         ``UserMessage`` object and stored in the conversation tracker.

        Args:
            request: incoming request with the message of the user

        Returns:
            Metadata which was extracted from the request.
        """
        pass
    
    def handle_message(self, message: Dict[Text, Any]) -> Message:
        """ handle client message """
        if message.get("type") == "text":
            return self.handle_text_message(message.pop("sender_id"), message.pop("text"), **message)
        
        if message.get("type") == "image":
            return self.handle_image_message(message.pop("sender_id"), message.pop("image"), **message)

    def handle_text_message(self, sender_id:Text, text:Text, **kwargs: Any) -> Message:
        return Message(
            sender = 'user',
            sender_id = sender_id,
            receiver = kwargs.get("receiver", None),
            receiver_id = kwargs.get("receiver_id", None),
            session_id = kwargs.get("session_id", None),
            text = text,
            metadata = {}
        )

    def handle_image_message(self, sender_id:Text, image:Text, **kwargs: Any) -> Message:
        return Message(
            sender = 'user',
            sender_id = sender_id,
            receiver = kwargs.get("receiver", None),
            receiver_id = kwargs.get("receiver_id", None),
            session_id = kwargs.get("session_id", None),
            text = "",
            metadata = {"image": image}
        )

    async def send_response(self, recipient_id: Text, message: Dict[Text, Any]) -> None:
        """Send a message to the client."""
        messages = convert_message_dict(message)

        for message in messages:
            if message.get('type') == 'text':
                await self.send_text_message(recipient_id, **message)
            
            if message.get('type') == 'image':
                await self.send_image_url(recipient_id, **message)

    async def send_text_message(
        self, recipient_id: Text, **kwargs: Any
    ) -> None:
        """Send a message through this channel."""

        raise NotImplementedError(
            "Output channel needs to implement a send message for simple texts."
        )

    async def send_image_url(
        self, recipient_id: Text, **kwargs: Any
    ) -> None:
        """Sends an image. Default will just post the url as a string."""

        await self.send_text_message(recipient_id, f"Image: {image}")


import logging
import uuid
import json
import time
import asyncio
from typing import Any, Awaitable, Callable, Dict, List, Optional, Text, Set
from sanic import Blueprint, Sanic
from sanic.request import Request
from sanic.response import HTTPResponse
from websockets import WebSocketServerProtocol, ConnectionClosed

from cota.channels.channel import Channel
from cota.message.message import Message
from cota.channels.connection import ConnectionPool
from cota.channels.room import RoomManager

logger = logging.getLogger(__name__)


class Websocket(Channel):
    """A websocket channel with room management."""
    
    @classmethod
    def name(cls) -> Text:
        return "websocket"
    
    def __init__(
        self,
        namespace: Optional[Text] = None,
        websocket_path: Optional[Text] = "/websocket",
        connection_timeout: int = 1000,
        room_timeout: int = 3600,
    ):
        self.namespace = namespace
        self.websocket_path = websocket_path
        
        # Use connection pool and room manager
        self._connection_pool = ConnectionPool()
        self._room_manager = RoomManager()
                
        self._connection_timeout = connection_timeout
        self._room_timeout = room_timeout
        
        # Cleanup task
        self._cleanup_task = None
    
    def blueprint(
        self, on_new_message: Callable[[Message], Awaitable[Any]]
    ) -> Blueprint:
        bp = Blueprint('websocket')

        # Start background tasks when the server starts
        @bp.before_server_start
        async def before_server_start(app: Sanic):
            if self._cleanup_task is None:
                self._cleanup_task = app.add_task(self._cleanup_loop())
                logger.info("Started WebSocket cleanup background task")

        # Handle WebSocket connections
        @bp.websocket(self.websocket_path)
        async def ws_route(request: Request, ws: WebSocketServerProtocol):
            sender_id = request.args.get("sender_id", None)
            if not sender_id:
                await ws.close(4001, "No sender_id provided")
                return

            sender_id = str(sender_id)
            await self._connection_pool.add(sender_id, ws)
            
            try:
                # send login confirmation
                await ws.send(json.dumps({
                    "type": "login", 
                    "sender_id": sender_id,
                    "timestamp": time.time()
                }))
                
                await self.on_connect(ws, on_new_message)
                    
            except ConnectionClosed as e:
                logger.info(f"WebSocket connection closed by client: {e}")
            except Exception as e:
                logger.error(f"Error in WebSocket connection: {e}", exc_info=True)
                await ws.close(1011, "Internal server error")
            finally:
                await self._connection_pool.remove(sender_id)
        
        return bp
    
    async def _cleanup_loop(self):
        """Periodically clean up inactive connections and rooms"""
        while True:
            try:
                await asyncio.sleep(300)  # cleanup every 5 minutes
                
                conn_count = await self._connection_pool.cleanup_inactive(
                    self._connection_timeout
                )
                if conn_count > 0:
                    logger.info(f"Cleaned up {conn_count} inactive connections")
                
                room_count = await self._room_manager.cleanup_inactive(
                    self._room_timeout
                )
                if room_count > 0:
                    logger.info(f"Cleaned up {room_count} inactive rooms")
                
            except asyncio.CancelledError:
                logger.info("WebSocket cleanup task cancelled")
                break
            except Exception as e:
                logger.error(f"Cleanup error: {e}", exc_info=True)
                await asyncio.sleep(5)
    
    async def on_connect(self, ws, on_new_message):
        """Handle WebSocket connections and messages"""
        try:
            while True:
                try:
                    message = await ws.recv()
                    msg = json.loads(message)

                    if not self._validate_message(msg):
                        logger.warning(f"Invalid message format: {msg}")
                        continue
                    
                    # get message type, session_id, sender_id, room_id
                    msg_type = msg.get('type')
                    session_id = msg.get("session_id")
                    sender_id = msg.get("sender_id")
                    room_id = session_id

                    if msg_type == 'join':
                        await self.on_join(msg, ws)
                        continue
                    elif msg_type == 'ping':
                        # update connection activity, should ensure client send ping message
                        await self._connection_pool.update_activity(sender_id)
                        await ws.send(json.dumps({
                            "type": "pong", 
                            "timestamp": time.time()
                        }))
                        continue

                    message_obj = self.handle_message(msg)
                    await on_new_message(message_obj, self)
                    
                except json.JSONDecodeError as e:
                    logger.warning(f"Failed to parse message: {e}")
                except ConnectionClosed:
                    break
                except Exception as e:
                    logger.error(f"Error processing message: {e}", exc_info=True)
        
        except Exception as e:
            logger.error(f"Unhandled exception in on_connect: {e}", exc_info=True)
    
    def _validate_message(self, msg):
        """Validate message format"""
        required_fields = ["session_id", "sender_id"]
        return all(field in msg for field in required_fields)
    
    async def on_join(self, msg, ws):
        """Handle join room request"""
        session_id = msg.get("session_id")
        sender_id = msg.get("sender_id")
        
        if not session_id or not sender_id:
            return
            
        try:
            room_id = session_id
            # add user to room
            await self._room_manager.add_user_to_room(room_id, sender_id)
            
            # get room users
            room_users = await self._room_manager.get_room_users(room_id)

            # send join confirm
            await ws.send(json.dumps({
                "type": "join_confirm",
                "session_id": session_id,
                "room_id": room_id,
                "timestamp": time.time(),
                "room_info": {
                    "user_count": len(room_users),
                    "users": room_users,
                    "room_id": room_id
                }
            }))

            # broadcast user joined
            await self.broadcast_to_room(
                room_id,
                {
                    "type": "user_joined",
                    "session_id": session_id,
                    "room_id": room_id,
                    "user_id": sender_id,
                    "timestamp": time.time(),
                    "room_info": {
                        "user_count": len(room_users),
                        "users": room_users,
                        "room_id": room_id
                    }
                },
                exclude_user=sender_id
            )
        except Exception as e:
            logger.error(f"Error handling join request: {e}")
            await ws.send(json.dumps({
                "type": "join_error",
                "session_id": session_id,
                "timestamp": time.time(),
                "error": str(e)
            }))
    
    async def send_text_message(self, recipient_id: Text, **kwargs: Any) -> None:
        """Send text message"""
        await self._send_message(recipient_id, kwargs)
    
    async def send_image_url(self, recipient_id: Text, **kwargs: Any) -> None:
        """Send image URL"""
        await self._send_message(recipient_id, kwargs)
    
    async def _send_message(self, recipient_id: Text, response: Any) -> None:
        """Send message to recipient(s)"""
        user_ids = await self._room_manager.get_room_users(str(recipient_id))
        
        sent_count = 0
        for user_id in user_ids:
            ws = await self._connection_pool.get(str(user_id))
            if ws:
                try:
                    await ws.send(json.dumps(response))
                    sent_count += 1
                except Exception as e:
                    logger.error(f"Error sending message to {user_id}: {e}")
        
        if sent_count == 0:
            logger.warning(f"Failed to send message to any recipient in room {recipient_id}")
    
    async def broadcast(self, message: Dict[str, Any], exclude_user: Optional[str] = None) -> int:
        """Broadcast message to all connected users"""
        connections = await self._connection_pool.get_all_connections()
        sent_count = 0
        
        for user_id, ws in connections.items():
            if user_id != exclude_user:
                try:
                    await ws.send(json.dumps(message))
                    sent_count += 1
                except Exception as e:
                    logger.error(f"Error broadcasting to {user_id}: {e}")
        
        return sent_count
    
    async def broadcast_to_room(self, room_id: str, message: Dict[str, Any], exclude_user: Optional[str] = None) -> int:
        """Broadcast message to all users in the room"""
        user_ids = await self._room_manager.get_room_users(str(room_id))
        sent_count = 0
        
        for user_id in user_ids:
            if user_id != exclude_user:
                ws = await self._connection_pool.get(str(user_id))
                if ws:
                    try:
                        await ws.send(json.dumps(message))
                        sent_count += 1
                    except Exception as e:
                        logger.error(f"Error broadcasting to room member {user_id}: {e}")
        
        return sent_count
    
    def handle_message(self, message: Dict[str, Any]) -> Message:
        """Convert original message to Message object"""
        return Message(message)


import logging
from typing import Optional, Dict
from .client import HttpClient, HttpConfig

logger = logging.getLogger(__name__)

class HttpClientManager:
    """HTTP client manager
    
    Unified management of HTTP client instances, implementing client reuse and lifecycle management
    """
    _instance = None
    _clients: Dict[str, HttpClient] = {}
    
    def __init__(self):
        if HttpClientManager._instance is not None:
            raise RuntimeError("HttpClientManager is a singleton!")
        HttpClientManager._instance = self
    
    @classmethod
    def instance(cls) -> 'HttpClientManager':
        """Get singleton instance"""
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance
    
    async def get_client(self, config_key: str = "default", config: Optional[HttpConfig] = None) -> HttpClient:
        """Get or create HTTP client
        
        Args:
            config_key: Configuration key, used to reuse clients with same configuration
            config: HTTP client configuration, create new client if not exists
            
        Returns:
            HttpClient instance
        """
        if config_key not in self._clients:
            client = HttpClient(config or HttpConfig())
            await client.connect()
            self._clients[config_key] = client
            
        return self._clients[config_key]
    
    async def close_all(self):
        """Close all client connections"""
        for client in self._clients.values():
            await client.close()
        self._clients.clear()
    
    async def close_client(self, config_key: str):
        """Close specified client connection"""
        if config_key in self._clients:
            await self._clients[config_key].close()
            del self._clients[config_key] 


import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Text, List, Union, Optional, Dict, Any

logger = logging.getLogger(__name__)


class Knowledge(ABC):
    """Base class for knowledge retrieval systems."""
    
    def __init__(self, config: Dict[str, Any] = None):
        self.config = config or {}
    
    @abstractmethod
    async def retrieve(self, query: str, context: Dict[str, Any] = None) -> str:
        """Retrieve knowledge based on query and context.
        
        Args:
            query: The search query
            context: Additional context information (DST, action, etc.)
            
        Returns:
            Retrieved knowledge as string
        """
        pass
    
    async def process_query(self, query: str, context: Dict[str, Any] = None) -> str:
        """Process query and return formatted result."""
        try:
            result = await self.retrieve(query, context)
            return self._format_result(result)
        except Exception as e:
            logger.error(f"Knowledge retrieval failed: {e}")
            return ""
    
    def _format_result(self, result: str) -> str:
        """Format the retrieval result."""
        return result.strip() if result else ""


class CompositeKnowledge(Knowledge):
    """
    Composite Knowledge - manages multiple Knowledge strategies internally.
    Implements true factory pattern: returns single object with internal strategy composition.
    """
    
    def __init__(self, strategies: List[Knowledge]):
        super().__init__()
        self.strategies = strategies
        logger.info(f"CompositeKnowledge initialized with {len(strategies)} strategies: "
                   f"{[s.__class__.__name__ for s in strategies]}")
    
    async def retrieve(self, query: str, context: Dict[str, Any] = None) -> str:
        """Iterate through strategies for knowledge retrieval, return first non-empty result."""
        for strategy in self.strategies:
            try:
                result = await strategy.retrieve(query, context)
                if result and result.strip():
                    logger.debug(f"Knowledge retrieved by {strategy.__class__.__name__}")
                    return result
            except Exception as e:
                logger.warning(f"Strategy {strategy.__class__.__name__} failed: {e}")
                continue
        return ""
    
    def add_strategy(self, strategy: Knowledge) -> None:
        """Dynamically add strategy."""
        self.strategies.append(strategy)
    
    def remove_strategy(self, strategy_class: type) -> bool:
        """Remove strategy of specified type."""
        for i, strategy in enumerate(self.strategies):
            if isinstance(strategy, strategy_class):
                del self.strategies[i]
                return True
        return False


class KnowledgeFactory:
    """Factory for creating Knowledge instances."""
    
    @staticmethod
    def create(knowledge_configs: List[Dict[str, Any]], path: Text = None) -> Optional[Knowledge]:
        """Create single Knowledge object based on configuration.
        Returns CompositeKnowledge for multiple configs, single Knowledge for one config,
        or None if no configurations provided.
        
        Args:
            knowledge_configs: List of knowledge configuration dictionaries (can be empty)
            path: Base path for knowledge resources
            
        Returns:
            Optional[Knowledge]: Single Knowledge object, CompositeKnowledge, or None if no valid config
        """
        if not knowledge_configs:
            logger.info("No knowledge configurations provided, knowledge will be disabled")
            return None
        
        # If only one config, return single Knowledge instance
        if len(knowledge_configs) == 1:
            config = knowledge_configs[0]
            try:
                return KnowledgeFactory._create_single_knowledge(config, path)
            except Exception as e:
                logger.info("Knowledge will be disabled due to configuration error")
                return None
        
        # Multiple configs: create CompositeKnowledge
        strategies = []
        for config in knowledge_configs:
            try:
                strategy = KnowledgeFactory._create_single_knowledge(config, path)
                strategies.append(strategy)
            except Exception as e:
                logger.error(f"Failed to create knowledge strategy from config {config}: {e}")
                continue
        
        if not strategies:
            logger.warning("Failed to create any knowledge strategies, knowledge will be disabled")
            return None
        
        return CompositeKnowledge(strategies)
    
    @staticmethod
    def _create_single_knowledge(config: Dict[str, Any], path: Text = None) -> Knowledge:
        """Create single Knowledge strategy instance."""
        if not config or not isinstance(config, dict):
            raise ValueError(f"Invalid knowledge config (not a dict): {config}")
            
        knowledge_type = config.get('type')
        
        if not knowledge_type:
            raise ValueError(f"Invalid knowledge config (missing type): {config}")
        
        if knowledge_type == 'llm':
            try:
                from cota.knowledge.llm_knowledge import LLMKnowledge
                llm_config = config.get('config', config)
                if not llm_config:
                    raise ValueError("LLM knowledge config is empty")
                return LLMKnowledge(config=llm_config)
            except ImportError as e:
                raise ValueError(f"LLMKnowledge not available: {e}")
            except Exception as e:
                raise ValueError(f"Failed to create LLMKnowledge: {e}")
        else:
            raise ValueError(f"Unknown knowledge type: {knowledge_type}")



from abc import ABC, abstractmethod
from typing import Dict, Text, Tuple, Any

class BasePlugin(ABC):
    """Plugin base class
    
    All user-defined plugins should inherit from this class and implement necessary methods.
    """
    
    def __init__(self, config: Dict[Text, Any]) -> None:
        """Initialize plugin
        
        Args:
            config: Plugin configuration
        """
        self.config = config
        
    @abstractmethod
    async def execute(self, data: Dict[Text, Any]) -> Tuple[Text, Dict]:
        """Execute plugin logic
        
        Args:
            data: Data required for execution
            
        Returns:
            Tuple[Response text, Metadata]
        """
        pass
        
    async def cleanup(self) -> None:
        """Clean up plugin resources
        
        If plugin needs to clean up resources when closing, can override this method.
        """
        pass 


"""LLM client factory for creating appropriate client instances."""

from typing import Dict

from .base import LLMClient
from .openai import OpenAIClient, OpenAIRAGClient
from .custom import CustomHttpClient


class LLMClientFactory:
    """Factory class for creating LLM client instances based on configuration."""
    
    @staticmethod
    def create_client(config: Dict) -> LLMClient:
        """Create appropriate LLM client based on configuration.
        
        Args:
            config: Configuration dictionary containing client type and parameters
            
        Returns:
            LLMClient instance
            
        Raises:
            ValueError: If client type is not supported
        """
        client_type = config.get('type', 'openai')
        
        # Extract common parameters
        api_key = config['key']
        base_url = config['apibase']
        model = config['model']
        
        if client_type == 'openai':
            return OpenAIClient(
                api_key=api_key,
                base_url=base_url,
                model=model
            )
        elif client_type == 'openai-rag':
            knowledge_id = config['knowledge_id']
            rag_prompt = config.get('rag_prompt')
            return OpenAIRAGClient(
                api_key=api_key,
                base_url=base_url,
                model=model,
                knowledge_id=knowledge_id,
                rag_prompt=rag_prompt
            )
        elif client_type == 'custom':
            extra_config = {
                k: v for k, v in config.items() 
                if k not in ['type', 'key', 'apibase', 'model']
            }
            return CustomHttpClient(
                api_key=api_key,
                base_url=base_url,
                model=model,
                **extra_config
            )
        else:
            raise ValueError(f"Unsupported client type: {client_type}")



import logging
import json
from cota.actions.action import Action
from cota.message.message import Message
from typing import Text, Optional, Dict, List, Any
from cota.dst import DST
from cota.constant import (
    DEFAULT_DIALOGUE_MAX_TOKENS
)

logger = logging.getLogger(__name__)

class UserUtter(Action):
    """User utterance action for processing user requests"""

    def apply_to(self, dst: DST) -> None:
        """Apply user utterance to dialogue state tracker"""
        dst.actions.append(self)
        dst.formless_actions.append(self)
        dst.latest_action = self
        dst.latest_query = self
        dst.latest_sender_id = self.sender_id
        dst.latest_receiver_id = self.receiver_id

    def run_from_string(self, text: Text) -> None:
        """Create user message from text string"""
        message = Message(
            sender='user',
            text=text
        )
        self.result.append(message.as_dict())

    async def run(
            self,
            agent: Optional["Agent"] = None,
            dst: Optional[DST] = None,
            user: Optional[Dict] = None,
    ):
        """Execute user utterance action with LLM processing"""
        knowledge_dict = await dst.format_knowledge(self.prompt, self)
        thoughts_dict = await dst.format_policies(self.prompt, self)
        prompt = dst.format_prompt(self.prompt, self, {**knowledge_dict, **thoughts_dict})

        if user and user.get('description'):
            messages = [{"role": "system", "content": user.get('description')},{"role":"user", "content": prompt}]
        else:
            messages = [{"role":"user", "content": prompt}]

        result = await agent.llm_instance(self.llm).generate_chat(
            messages = messages,
            max_tokens = agent.dialogue.get('max_tokens', DEFAULT_DIALOGUE_MAX_TOKENS),
            response_format={'type': 'json_object'}
        )

        # Parse JSON response from LLM
        content = result["content"]
        try:
            json_result = json.loads(content)
            text_content = json_result.get('text', '')
            thought_content = json_result.get('thought', '')
            state_content = json_result.get('state', 'continue')
            
            logger.debug(f"Parsed JSON result: {json_result}")
        except json.JSONDecodeError:
            logger.error(f"Failed to parse JSON response from user: {content}")
            # Fallback to original text if JSON parsing fails
            text_content = content
            thought_content = ''
            state_content = 'continue'

        # Create message with text content
        message = Message(
            sender='user',
            text=text_content
        )
        
        # Create message dict and add thought and state at the same level as text
        message_dict = message.as_dict()
        if thought_content:
            message_dict['thought'] = thought_content
        if state_content:
            message_dict['state'] = state_content
            
        self.result.append(message_dict)
        
        logger.debug(f"Query proxy user: {user}")
        logger.debug(f"Query prompt: {prompt}")
        logger.debug(f"Query result: {self.result}")



import logging
import asyncio
from typing import Dict, Text, Tuple, Any
from .base import Executor

logger = logging.getLogger(__name__)

class PythonExecutor(Executor):
    """Python executor
    
    Handles Python code execution logic. Supports both synchronous and asynchronous code execution.
    """
    
    async def execute(self, data: Dict[Text, Any]) -> Tuple[Text, Dict]:
        """Execute Python code
        
        Args:
            data: Data required for execution
            
        Returns:
            Tuple[Response text, Metadata]
        """
        code = self.config.get("code", "")
        if not code:
            return "", {"error": "No code provided"}
            
        try:
            # Create local namespace
            namespace = {
                "data": data,
                "config": self.config,
                "result": {"text": "", "metadata": {}}
            }
            
            # Check if code is an async function
            is_async = "async def" in code
            
            if is_async:
                # If it's async code, use exec to compile and execute
                exec_code = f"""
async def _execute():
    {code}
"""
                exec(exec_code, namespace)
                await namespace["_execute"]()
            else:
                # If it's sync code, execute directly
                exec(code, namespace)
            
            result = namespace.get("result", {})
            return result.get("text", ""), result.get("metadata", {})
            
        except Exception as e:
            logger.error(f"Failed to execute Python code: {str(e)}")
            return "", {"error": str(e)} 


"""LLM (Large Language Model) module for COTA.

This module provides a unified interface for interacting with various
language model providers including OpenAI-compatible APIs.
"""

from .llm import LLM
from .base import LLMClient
from .factory import LLMClientFactory
from .openai import OpenAIClient, OpenAIRAGClient
from .custom import CustomHttpClient

__all__ = [
    'LLM',
    'LLMClient', 
    'LLMClientFactory',
    'OpenAIClient',
    'OpenAIRAGClient', 
    'CustomHttpClient'
]



import sys
import logging
import uuid
import json
from typing import Any, Awaitable, Callable, Dict, Iterable, List, Optional, Text
from sanic import Blueprint, response, Sanic
from sanic.request import Request
from sanic.response import HTTPResponse
import websockets
from cota.channels.channel import Channel
from cota.message.message import Message

logger = logging.getLogger(__name__)


class Cmdline(Channel):
    """A cmdline channel."""
    @classmethod
    def name(cls) -> Text:
        return "cmdline"

    def __init__(
        self,
        namespace: Optional[Text] = None,
        on_new_message: Callable[[Message], Awaitable[Any]] = None
    ):
        self.namespace = namespace
        self.on_new_message = on_new_message

    def blueprint(
        self, on_new_message: Callable[[Message], Awaitable[Any]]
    ) -> Blueprint:
        pass

    async def on_connect(self):
        """Handle the connection and continuously read messages from the command line."""
        session_id = uuid.uuid4().hex
        while True:
            print("Input message:")
            message = input().strip()

            if message == "/stop":
                sys.exit(0)

            message = {
                    'type': 'text',
                    'sender': 'user',
                    'sender_id': 'default_user',
                    'receiver': 'bot',
                    'receiver_id': 'default_bot',
                    'session_id': session_id,
                    'text': message,
                    'metadata': {}
                }
            message = self.handle_message(message)
            await self.on_new_message(message, self)

    async def send_text_message(
        self, recipient_id: Text, **kwargs: Any
    ) -> None:
        """Send a message through this channel."""
        await self._send_message(recipient_id, kwargs)

    async def send_image_url(
        self, recipient_id: Text, **kwargs: Any
    ) -> None:
        """Sends an image to the output"""
        await self._send_message(recipient_id, kwargs)

    async def _send_message(self, recipient_id: Text, response: Any) -> None:
        """Sends a message to the recipient using the bot event."""
        sender = response.get('sender', 'unknown')
        text = response.get('text', '')
        
        if sender == 'user':
            print('User: ', text)
        elif sender == 'bot':
            print('Bot: ', text)
        else:
            print(f'{sender.title()}: ', text)



import logging
import time
import json
from typing import Text, Optional, List, Dict, Any, Tuple
from cota.actions.action import Action
from cota.dst import DST
from cota.utils.parser import extract_json_from_string
from cota.utils.common import all_keys_filled, update_existing_keys
from cota.utils.http import HttpClientManager, HttpConfig
from cota.constant import (
    DEFAULT_FORM_UPDATER_INSTRUCTION,
    DEFAULT_DIALOGUE_MAX_TOKENS
)

logger = logging.getLogger(__name__)

class Form(Action):

    def __init__(
            self,
            name: Optional[Text] = None,
            description: Optional[Text] = None,
            prompt: Optional[Text] = None,
            llm: Optional[Text] = None,
            timestamp: Optional[float] = None,
            sender_id: Optional[Text] = None,
            receiver_id: Optional[Text] = None,
            result: Optional[List] = None,
            slots: Optional[Dict] = None,
            state: Optional[Text] = None,
            executer: Optional[Dict] = None
    ) -> None:
        super().__init__(
            name=name,
            description=description,
            prompt=prompt,
            llm=llm,
            timestamp=timestamp,
            sender_id=sender_id,
            receiver_id=receiver_id,
            result=result        
        )
        self.slots = slots or {}
        self.state = state
        self.executer = executer or {}

    @staticmethod
    def resolve_by_type(
            name: Text
    ) -> "Form":
        new_class = type(name, (Form,), {})
        new_class.__name__ = name
        return new_class

    @staticmethod
    def build_from_name(**kwargs) -> "Form":
        action_class = Form.resolve_by_type(kwargs.get("name"))
        if not action_class:
            return None
        return action_class(**kwargs)

    def apply_to(self, dst: DST) -> None:
        """
        Only add the action to the list when it is at the beginning or end
        """
        dst.current_form = self if self.state in ("start", "continue") else None

        dst.actions.append(self)
        if self.state == 'end':
            dst.formless_actions.append(self)

        # Update latest action
        dst.latest_action = self

    def run_from_dict(
            self,
            action_dict
    ) -> None:
        self.name = action_dict.get("name") or self.name
        self.timestamp = action_dict.get("timestamp") or self.timestamp
        self.sender_id = action_dict.get("sender_id") or self.sender_id
        self.receiver_id = action_dict.get("receiver_id") or self.receiver_id
        self.slots = action_dict.get("slots") or self.slots
        self.state = action_dict.get("state") or self.state
        self.result = action_dict.get("result")

    def run_from_string(self, text: Text) -> None:
        self.result.append(
            {
                "action_name": text,
            }
        )

    def as_dict(self) -> Dict[Text, Any]:
        d = {
            "name": self.__class__.__name__,
            "timestamp": self.timestamp,
            "sender_id": self.sender_id,
            "receiver_id": self.receiver_id,
            "slots": self.slots,
            "state": self.state
        }
        if self.result:
            d["result"] = self.result
        return d

    async def update(self, agent, dst):
        prompt = None
        update_result = None
        
        try:
            prompt_template = self.prompt
            if not prompt_template:
                return
            knowledge_dict = await dst.format_knowledge(prompt_template, self)
            thoughts_dict = await dst.format_policies(prompt_template, self)
            prompt = dst.format_prompt(prompt_template, self, {**knowledge_dict, **thoughts_dict})

            update_result = await agent.llm_instance(self.llm).generate_chat(
                messages = [{"role": "system", "content": DEFAULT_FORM_UPDATER_INSTRUCTION},{"role":"user", "content": prompt}],
                max_tokens = agent.dialogue.get('max_tokens', DEFAULT_DIALOGUE_MAX_TOKENS),
                response_format = {'type': 'json_object'}
            )
            # Extract content from result
            content = update_result["content"]
            try:
                update_json = json.loads(content)
                if len(update_json) > 0:
                    update_existing_keys(self.slots, update_json)
            except json.JSONDecodeError:
                logger.error(f"Failed to parse JSON response from {self.name} slot updater: {content}")
                try:
                    update_json = extract_json_from_string(content)
                    if len(update_json) > 0:
                        update_existing_keys(self.slots, update_json)
                except Exception as e:
                    logger.error(f"Failed to parse JSON response from {self.name} slot updater: {e}")
        except Exception as e:
            logger.error(f"Failed to update {self.name}: {e}")

        logger.debug( f"{self.name} slot update prompt: {prompt}" )
        logger.debug( f"{self.name} slot update result: {update_result}" )
        logger.debug( f"{self.name} slots after update: {self.slots}")

    async def execute(self, agent, data: Dict[Text, Any]) -> Tuple[Text, Dict]:
        """Execute form
        
        Args:
            data: Data required for execution
            
        Returns:
            Tuple[Response text, Metadata]
        """
        # First check if it's a mock execution
        if self.executer.get("mock", False):
            logger.info(f"Mock execution for {self.name}")
            output_options = self.executer.get("output", [])
            return await self._handle_mock_execution(output_options)
            
        # Get executor
        executor = agent.get_executor(self.name)
        if not executor:
            return "", {"error": f"No executor found for form: {self.name}"}
            
        return await executor.execute(data)
        
    async def _handle_mock_execution(self, data: List[Text]) -> Tuple[Text, Dict]:
        """Handle mock execution
        
        Args:
            data: Optional list of output texts
            
        Returns:
            Tuple[Response text, Metadata]
        """
        if not data:
            return "", {"error": "No mock outputs available"}
            
        # Display all available output options
        print("\nAvailable mock outputs:")
        for i, text in enumerate(data, 1):
            print(f"{i}. {text}")
            
        # Let user choose
        while True:
            try:
                choice = int(input("\nPlease select an output option (enter number): "))
                if 1 <= choice <= len(data):
                    break
                print("Invalid choice. Please enter a valid number.")
            except ValueError:
                print("Please enter a valid number.")
                
        # Get selected output
        output_text = data[choice - 1]
        
        # If output contains <text>, let user input specific text
        if "<text>" in output_text:
            user_text = input("\nPlease enter your text: ")
            output_text = output_text.replace("<text>", user_text)
                
        return output_text, {}

    async def run(
            self,
            agent: Optional["Agent"] = None,
            dst: Optional[DST] = None
    ) -> List[Action]:
        text = ""
        metadata = None

        # handle init state
        if not self.state:
            self.state = "start"
            update_existing_keys(self.slots, dst.slots)

            # try to update slots if needed
            if not all_keys_filled(self.slots):
                await self.update(agent, dst)

        # handle continue state
        elif self.state == "start" or self.state == "continue":
            # try to update slots if needed
            if not all_keys_filled(self.slots):
                await self.update(agent, dst)

            self.state = "continue"
            # check if slots are now filled
            if all_keys_filled(self.slots):
                self.state = "end"
                text, metadata = await self.execute(agent, self.slots)


        self.result.append({"text": text, "metadata": metadata})
        logger.debug( f"{self.name} result: {self.result}" )